import configparser
from pathlib import Path

import pyramid.config.settings


def apply_local_settings(settings: pyramid.config.settings.Settings) -> None:
    """Apply to the ``settings`` dict the settings found in the local settings file."""
    local_settings_path_str = settings.get("local_settings_path")
    local_settings_path = None if local_settings_path_str is None else Path(local_settings_path_str)
    if local_settings_path is not None and local_settings_path.exists():
        config_parser = configparser.ConfigParser()
        config_parser.read(local_settings_path_str)
        settings.update(config_parser.items("app:main"))



import json

import pytest
from colander import Invalid, null
from geoalchemy2 import WKBElement
from geoalchemy2.shape import from_shape, to_shape

from c2cgeoform.tests import DatabaseTestCase


class TestGeometry(DatabaseTestCase):
    def test_serialize_null(self):
        from c2cgeoform.ext.colander_ext import Geometry

        geom_schema = Geometry()

        assert null == geom_schema.serialize({}, null)

    def test_serialize_wkb(self):
        from c2cgeoform.ext.colander_ext import Geometry

        geom_schema = Geometry()

        from shapely.geometry.point import Point

        wkb = from_shape(Point(1.0, 2.0))
        assert {"type": "Point", "coordinates": [1.0, 2.0]} == json.loads(geom_schema.serialize({}, wkb))

    def test_serialize_reproject(self):
        from c2cgeoform.ext.colander_ext import Geometry

        geom_schema = Geometry(srid=4326, map_srid=3857)

        from shapely.geometry.point import Point

        wkb = from_shape(Point(1.0, 2.0), 4326)
        geo_json = json.loads(geom_schema.serialize({}, wkb))
        assert "Point" == geo_json["type"]
        assert round(geo_json["coordinates"][0], 7) == round(111319.49079327231, 7)
        assert round(geo_json["coordinates"][1], 7) == round(222684.20850554455, 7)

    def test_serialize_invalid(self):
        from c2cgeoform.ext.colander_ext import Geometry

        geom_schema = Geometry()

        with pytest.raises(Invalid):
            geom_schema.serialize({}, "Point(1 0)")

    def test_deserialize_null(self):
        from c2cgeoform.ext.colander_ext import Geometry

        geom_schema = Geometry()

        assert null == geom_schema.deserialize({}, null)
        assert null == geom_schema.deserialize({}, "")

    def test_deserialize_valid_geojson(self):
        from c2cgeoform.ext.colander_ext import Geometry

        geom_schema = Geometry()

        from shapely.geometry.point import Point

        expected_wkb = WKBElement(Point(1.0, 2.0).wkb)

        wkb = geom_schema.deserialize({}, '{"type": "Point", "coordinates": [1.0, 2.0]}')
        assert expected_wkb.desc == wkb.desc

    def test_deserialize_reproject(self):
        from c2cgeoform.ext.colander_ext import Geometry

        geom_schema = Geometry(srid=4326, map_srid=3857)

        wkb = geom_schema.deserialize(
            {}, '{"type": "Point", "coordinates": [111319.49079327231, 222684.20850554455]}'
        )
        assert 4326 == wkb.srid

        shape = to_shape(wkb)
        assert round(shape.x, 7) == round(1.0, 7)
        assert round(shape.y, 7) == round(2.0, 7)

    def test_serialize_invalid_wrong_type(self):
        from c2cgeoform.ext.colander_ext import Geometry

        geom_schema = Geometry()

        with pytest.raises(Invalid):
            geom_schema.deserialize({}, '{"type": "InvalidType", "coordinates": [1.0, 2.0]}')

    def test_serialize_invalid_syntax(self):
        from c2cgeoform.ext.colander_ext import Geometry

        geom_schema = Geometry()

        with pytest.raises(Invalid):
            geom_schema.deserialize({}, '"type": "Point", "coordinates": [1.0, 2.0]}')


class TestBinaryData(DatabaseTestCase):
    def test_serialize_anything(self):
        from c2cgeoform.ext.colander_ext import BinaryData

        binary = BinaryData()
        serialized = binary.serialize({}, b"a string of binary data")
        assert null != serialized
        assert b"a string of binary data" == serialized.getvalue()

    def test_serialize_null(self):
        from c2cgeoform.ext.colander_ext import BinaryData

        binary = BinaryData()
        assert null == binary.serialize({}, null)

    def test_serialize_empty_string(self):
        from c2cgeoform.ext.colander_ext import BinaryData

        binary = BinaryData()
        assert null == binary.serialize({}, "")

    def test_deserialize_null(self):
        from c2cgeoform.ext.colander_ext import BinaryData

        binary = BinaryData()
        assert null == binary.deserialize({}, null)

    def test_deserialize_empty_string(self):
        from c2cgeoform.ext.colander_ext import BinaryData

        binary = BinaryData()
        assert null == binary.deserialize({}, "")

    def test_deserialize_file(self):
        import os

        from c2cgeoform.ext.colander_ext import BinaryData

        dirpath = os.path.dirname(os.path.realpath(__file__))
        file_ = open(os.path.join(dirpath, "data", "1x1.png"), "br")
        binary = BinaryData()
        assert isinstance(binary.deserialize({}, file_), bytes)

        # test that the file can be read multiple times (simulates that
        # a file in the tmpstore is requested several times)
        assert 95 == len(binary.deserialize({}, file_))
        assert 95 == len(binary.deserialize({}, file_))
        assert 95 == len(binary.deserialize({}, file_))



import json
import logging
import os
from collections.abc import Callable
from io import BufferedRandom, BytesIO
from typing import Any

import colander
import deform.field
import pyramid.interfaces
import pyramid.request
import requests
import sqlalchemy.orm
from colander import Invalid
from deform.compat import string_types
from deform.widget import (
    CheckboxChoiceWidget,
    MappingWidget,
    RadioChoiceWidget,
    Select2Widget,
    SelectWidget,
    Widget,
)
from deform.widget import FileUploadWidget as DeformFileUploadWidget
from sqlalchemy import inspect
from translationstring import TranslationString, TranslationStringFactory

from c2cgeoform import JSON, JSONDict, default_map_settings

_ = TranslationStringFactory("c2cgeoform")
log = logging.getLogger(__name__)


class MapWidget(Widget):  # type: ignore[misc]
    """
    Map Deform widget.

    A Deform widget that fits with GeoAlchemy 2 geometry columns and shows
    an OpenLayers map which allows to draw and modify geometries.

    Example usage

    .. code-block:: python

        geom = Column(
            geoalchemy2.Geometry('POLYGON', 4326), info={
                'colanderalchemy': {
                    'typ': colander_ext.Geometry(
                        'POLYGON', srid=4326, map_srid=3857),
                    'widget': deform_ext.MapWidget(
                        map_options={
                            'baseLayers': [{'type_': 'OSM'}],
                            'view':
                                'projection': 'EPSG:3857',
                                'center': [829170, 5933942],
                                'zoom': 7,
                            'fitMaxZoom': 14,
                            'focusOnly': False
                        }
                    )
                }})

    **Attributes/Arguments**

    map_options (dict, optional):
        Options for the map with:

        - baseLayers (list, optional): List of layer definitions.

        - view (dict, optional): View parameters (projection, center, zoom)

        - fitMaxZoom (int, optional): Maximum zoom when fitting to given geometry.

        - focusOnly (boolean, optional): If map has to be focused for interactions to work.

    If those parameters are not sufficient to customize the map,
    the template file `map.pt` can to be overwritten in application project.
    """

    requirements: tuple[tuple[str, str | None], ...] = ()

    map_options = default_map_settings

    def serialize(
        self,
        field: deform.field.Field,
        cstruct: str | colander._null,
        readonly: bool = False,
        **kw: Any,
    ) -> str:
        if cstruct is colander.null:
            cstruct = ""
        values = self.get_template_values(field, cstruct, kw)
        map_options = {
            key: (
                field.parent.schema.request.translate(value)
                if isinstance(value, TranslationString)
                else value
            )
            for key, value in self.map_options.items()
        }
        values["map_options"] = {**self._get_controls_definition(field, readonly), **map_options}
        # make `_` available in template for i18n messages
        values["_"] = _
        return field.renderer("map", **values)  # type: ignore[no-any-return]

    def deserialize(self, field: deform.field.Field, pstruct: str) -> str:
        del field  # Unused
        return pstruct

    def _get_controls_definition(self, field: deform.field.Field, readonly: bool) -> JSONDict:
        geometry_type = field.typ.geometry_type

        point = True
        line = True
        polygon = True
        is_multi_geometry = False

        if "POINT" in geometry_type:
            line = False
            polygon = False
        elif "LINESTRING" in geometry_type:
            point = False
            polygon = False
        elif "POLYGON" in geometry_type:
            point = False
            line = False

        if "MULTI" in geometry_type or geometry_type in ("GEOMETRY", "GEOMETRYCOLLECTION"):
            is_multi_geometry = True

        return {
            "point": point,
            "line": line,
            "polygon": polygon,
            "isMultiGeometry": is_multi_geometry,
            "readonly": readonly,
        }


class RelationSelectMixin:
    """Mixin class to support relations for select fields."""

    def __init__(
        self,
        model: type[Any],
        id_field: str,
        label_field: str,
        default_value: str | None = None,
        order_by: str | None = None,
    ) -> None:
        self.model = model
        self.id_field = id_field
        self.label_field = label_field
        self.default_value = default_value
        self.order_by = order_by

    def populate(self, session: sqlalchemy.orm.Session, request: pyramid.request.Request) -> None:
        del request  # unused
        self.values = self._get_select_values(session)

    def _get_select_values(self, session: sqlalchemy.orm.Session) -> tuple[str | tuple[str, str], ...]:
        model = inspect(self.model)
        order_by = getattr(model.columns, self.order_by) if self.order_by is not None else None

        entities = session.query(model).order_by(order_by)

        values: tuple[tuple[str, str], ...] = tuple(
            (getattr(entity, self.id_field), getattr(entity, self.label_field)) for entity in entities
        )

        if self.default_value is None:
            return values
        return (self.default_value, *values)


class RelationMultiSelectMixin(RelationSelectMixin):
    """Mixin class to support n:m relations for multi select fields."""

    def deserialize(self, field: deform.field.Field, pstruct: str) -> list[JSONDict]:
        """
        Deserialize the field input for a n:m relation.

        For example for a n:m relation between mapper A and mapper B and a
        relation table A_B: Let's assume this widget is used in model A, so
        that you can select entries of B. Then this method will receive a list
        of ids of table B. For each id it will create an object containing the
        id. These objects will be inserted in the relation table A_B.
        """
        if pstruct in (colander.null, None):
            return []
        ids = pstruct

        # get the id field of the mapped table B in the relation table A_B
        mapped_id_field = self._get_mapped_id_field(field)

        result = []
        if ids:
            assert mapped_id_field is not None
            for id_ in ids:
                # create an entry for the relation table A_B
                obj = {}
                # set the id for an entity of the mapped table B,
                # the id for an entity of table A will be filled in automatically
                obj[mapped_id_field.name] = mapped_id_field.deserialize(id_)
                result.append(obj)

        return result

    def serialize(
        self,
        field: deform.field.Field,
        cstruct: colander._null | str | None,
        **kw: Any,
    ) -> list[str]:
        """Flatten a list of objects into a list of ids."""
        del kw  # unused

        mapped_id_field = self._get_mapped_id_field(field)
        # create a list with only the ids of entities of table B
        if cstruct in (colander.null, None):
            cstruct = []
        assert cstruct is not None
        assert mapped_id_field is not None
        return [obj[mapped_id_field.name] for obj in cstruct]

    def _get_mapped_id_field(self, field: deform.field.Field) -> deform.field.Field | None:
        """
        Get the foreign key field of the mapped table B in the relation table A_B.

        For the given relation field in table A, find the foreign key field
        for table B in the relation table A_B.
        """
        relation_field = field.children[0]

        for subfield in relation_field.children:
            if subfield.name == self.id_field:
                return subfield
        return None


class RelationSelectWidget(SelectWidget, RelationMultiSelectMixin):  # type: ignore[misc]
    """
    Relation select Deform widget.

    Extension of the widget ````deform.widget.SelectWidget`` which loads the
    values from the database using a SQLAlchemy model.

    Example usage

    .. code-block:: python

        districtId = Column(Integer, ForeignKey('district.id'), info={
            'colanderalchemy': {
                'title': 'District',
                'widget': deform_ext.RelationSelectWidget(
                    District,
                    'id',
                    'name',
                    order_by='name',
                    default_value=('', _('- Select -'))
                )
            }})

    The values of this ``<select>`` field are filled with entries of model
    ``District``, whereas property ``id`` is used as value and ``name`` as
    label.

    For n:m relations the widget can be used like so:

    .. code-block:: python

        situations = relationship(
            "Situation",
            secondary=situation_for_permission,
            cascade="save-update,merge,refresh-expire",
            info={
                'colanderalchemy': {
                    'title': _('Situations'),
                    'widget': RelationSelectWidget(
                        Situation,
                        'id',
                        'name',
                        order_by='name',
                        multiple=True
                    ),
                    'includes': ['id'],
                    'validator': manytomany_validator
                }
            })

    **Attributes/Arguments**

    model (required)
        The SQLAlchemy model that is used to generate the list of values.

    id_field
        The property of the model that is used as value.
        Default: ``id``.

    label_field
        The property of the model that is used as label.
        Default: ``label``.

    order_by
        The property of the model that is used for the ``order_by`` clause of
        the SQL query.
        Default: ``None``.

    default_value
        A default value that is added add the beginning of the list of values
        that were loaded from the database.
        For example: ``default_value=('', _('- Select -'))``
        Default: ``None``.

    multiple
        Allow to select multiple values. Requires a n:m relationship.
        Default: ``False``.

    For further attributes, please refer to the documentation of
    ``deform.widget.SelectWidget`` in the deform documentation:
    <http://deform.readthedocs.org/en/latest/api.html>

    """

    def __init__(
        self,
        model: type[Any],
        id_field: str = "id",
        label_field: str = "label",
        default_value: Any = None,
        order_by: str | None = None,
        **kw: Any,
    ) -> None:
        RelationMultiSelectMixin.__init__(self, model, id_field, label_field, default_value, order_by)
        SelectWidget.__init__(self, **kw)

    def deserialize(self, field: deform.field.Field, pstruct: str) -> list[JSONDict]:
        if self.multiple:
            return RelationMultiSelectMixin.deserialize(self, field, pstruct)
        return SelectWidget.deserialize(self, field, pstruct)  # type: ignore[no-any-return]

    def serialize(self, field: deform.field.Field, cstruct: Any, **kw: Any) -> list[str]:
        cstruct_internal = (
            RelationMultiSelectMixin.serialize(self, field, cstruct, **kw) if self.multiple else cstruct
        )
        return SelectWidget.serialize(self, field, cstruct_internal, **kw)  # type: ignore[no-any-return]


class RelationSelect2Widget(Select2Widget, RelationMultiSelectMixin):  # type: ignore[misc]
    """
    Relation select Deform widget.

    Extension of the widget ````deform.widget.Select2Widget`` which loads the
    values from the database using a SQLAlchemy model.

    Example usage

    .. code-block:: python

        districtId = Column(Integer, ForeignKey('district.id'), info={
            'colanderalchemy': {
                'title': 'District',
                'widget': deform_ext.RelationSelect2Widget(
                    District,
                    'id',
                    'name',
                    order_by='name',
                    default_value=('', _('- Select -'))
                )
            }})

    The values of this ``<select>`` field are filled with entries of model
    ``District``, whereas property ``id`` is used as value and ``name`` as
    label.

    For n:m relations the widget can be used like so:

    .. code-block:: python

        situations = relationship(
            "Situation",
            secondary=situation_for_permission,
            cascade="save-update,merge,refresh-expire",
            info={
                'colanderalchemy': {
                    'title': _('Situations'),
                    'widget': RelationSelect2Widget(
                        Situation,
                        'id',
                        'name',
                        order_by='name',
                        multiple=True
                    ),
                    'includes': ['id'],
                    'validator': manytomany_validator
                }
            })

    **Attributes/Arguments**

    model (required)
        The SQLAlchemy model that is used to generate the list of values.

    id_field
        The property of the model that is used as value.
        Default: ``id``.

    label_field
        The property of the model that is used as label.
        Default: ``label``.

    order_by
        The property of the model that is used for the ``order_by`` clause of
        the SQL query.
        Default: ``None``.

    default_value
        A default value that is added add the beginning of the list of values
        that were loaded from the database.
        For example: ``default_value=('', _('- Select -'))``
        Default: ``None``.

    multiple
        Allow to select multiple values. Requires a n:m relationship.
        Default: ``False``.

    For further attributes, please refer to the documentation of
    ``deform.widget.Select2Widget`` in the deform documentation:
    <http://deform.readthedocs.org/en/latest/api.html>

    """

    def __init__(
        self,
        model: type[Any],
        id_field: str = "id",
        label_field: str = "label",
        default_value: Any = None,
        order_by: str | None = None,
        **kw: Any,
    ) -> None:
        RelationMultiSelectMixin.__init__(self, model, id_field, label_field, default_value, order_by)
        Select2Widget.__init__(self, **kw)

    def deserialize(self, field: deform.field.Field, pstruct: str) -> list[JSONDict]:
        if self.multiple:
            return RelationMultiSelectMixin.deserialize(self, field, pstruct)
        return Select2Widget.deserialize(self, field, pstruct)  # type: ignore[no-any-return]

    def serialize(self, field: deform.field.Field, cstruct: Any, **kw: Any) -> list[str]:
        cstruct_internal = (
            RelationMultiSelectMixin.serialize(self, field, cstruct, **kw) if self.multiple else cstruct
        )
        return Select2Widget.serialize(self, field, cstruct_internal, **kw)  # type: ignore[no-any-return]


class RelationCheckBoxListWidget(CheckboxChoiceWidget, RelationMultiSelectMixin):  # type: ignore[misc]
    """
    Relation checkbox list Deform widget.

    Extension of the widget ````deform.widget.CheckboxChoiceWidget`` which
    loads the values from the database using a SQLAlchemy model.

    For n:m relations the widget can be used like so:

    .. code-block:: python

        situations = relationship(
            "Situation",
            secondary=situation_for_permission,
            cascade="save-update,merge,refresh-expire",
            info={
                'colanderalchemy': {
                    'title': _('Situations'),
                    'widget': RelationCheckBoxListWidget(
                        Situation,
                        'id',
                        'name',
                        order_by='name',
                        edit_url=lambda request, value: request.route_url(
                                'c2cgeoform_item',
                                table='situations',
                                id=value
                                )
                    ),
                    'includes': ['id'],
                    'validator': manytomany_validator
                }
            })

    **Attributes/Arguments**

    model (required)
        The SQLAlchemy model that is used to generate the list of values.

    id_field
        The property of the model that is used as value.
        Default: ``id``.

    label_field
        The property of the model that is used as label.
        Default: ``label``.

    order_by
        The property of the model that is used for the ``order_by`` clause of
        the SQL query.
        Default: ``None``.
    edit_url (optional)
        a function taking request and value as parameter and returning
        an url to the corresponding resource.

    For further attributes, please refer to the documentation of
    ``deform.widget.Select2Widget`` in the deform documentation:
    <http://deform.readthedocs.org/en/latest/api.html>
    """

    def __init__(
        self,
        model: type[Any],
        id_field: str = "id",
        label_field: str = "label",
        order_by: str | None = None,
        **kw: Any,
    ) -> None:
        RelationMultiSelectMixin.__init__(self, model, id_field, label_field, None, order_by)
        CheckboxChoiceWidget.__init__(self, multiple=True, **kw)

    def deserialize(self, field: deform.field.Field, pstruct: str) -> list[JSONDict]:
        return RelationMultiSelectMixin.deserialize(self, field, pstruct)

    def serialize(self, field: deform.field.Field, cstruct: Any, **kw: Any) -> list[str]:
        cstruct_internal = RelationMultiSelectMixin.serialize(self, field, cstruct, **kw)
        return CheckboxChoiceWidget.serialize(self, field, cstruct_internal, **kw)  # type: ignore[no-any-return]


class RelationRadioChoiceWidget(RadioChoiceWidget, RelationSelectMixin):  # type: ignore[misc]
    """
    Relation radio choice Deform widget.

    Extension of the widget ````deform.widget.RadioChoiceWidget`` which loads
    the values from the database using a SQLAlchemy model.

    Example usage

    .. code-block:: python

        districtId = Column(Integer, ForeignKey('district.id'), info={
            'colanderalchemy': {
                'title': 'District',
                'widget': deform_ext.RelationRadioChoiceWidget(
                    District,
                    'id',
                    'name',
                    order_by='name'
                )
            }})

    The values of this field are filled with entries of model ``District``,
    whereas property ``id`` is used as value and ``name`` as label.

    **Attributes/Arguments**

    model (required)
        The SQLAlchemy model that is used to generate the list of values.

    id_field
        The property of the model that is used as value.
        Default: ``id``.

    label_field
        The property of the model that is used as label.
        Default: ``label``.

    order_by
        The property of the model that is used for the ``order_by`` clause of
        the SQL query.
        Default: ``None``.

    For further attributes, please refer to the documentation of
    ``deform.widget.RadioChoiceWidget`` in the deform documentation:
    <http://deform.readthedocs.org/en/latest/api.html>

    """

    def __init__(
        self,
        model: type[Any],
        id_field: str = "id",
        label_field: str = "label",
        order_by: str | None = None,
        **kw: Any,
    ) -> None:
        RelationSelectMixin.__init__(self, model, id_field, label_field, None, order_by)
        RadioChoiceWidget.__init__(self, **kw)


class FileUploadTempStore:
    def __init__(self, session: pyramid.interfaces.ISession) -> None:
        super().__init__()
        self.session = session

    def get(self, name: str, default: Any = None) -> Any:
        return self.deserialize(self.session.get(name, default))

    def __getitem__(self, name: str) -> Any:
        return self.deserialize(self.session[name])

    def __setitem__(self, name: str, value: Any) -> None:
        self.session[name] = self.serialize(value)
        self.session.save()

    def __contains__(self, name: str) -> bool:
        return name in self.session

    def serialize(self, data: str) -> str:
        if isinstance(data, dict):
            return {k: self.serialize(v) for k, v in data.items()}
        if isinstance(data, BufferedRandom | BytesIO):
            value = data.read()
            # set the file position back to 0, so that the file can be read again
            data.seek(0, os.SEEK_SET)
            return value
        return data

    def deserialize(self, data: str) -> str:
        if isinstance(data, dict):
            return {k: self.deserialize(v) for k, v in data.items()}
        if isinstance(data, bytes):
            return BytesIO(data)
        return data

    def preview_url(self, name: str) -> None:
        pass


class FileUploadWidget(DeformFileUploadWidget):  # type: ignore[misc]
    """
    File upload Deform widget.

    Extension of ``deform.widget.FileUploadWidget`` to be used in a model
    class that extends the ``models.FileData`` mixin class.

    Note that, contrary to ``deform.widget.FileUploadWidget``, this extension
    is not meant to be used with the ``deform.FileData`` Colander type. Instead
    it works with the ``colander.Mapping`` type, which is what
    ``colanderalchemy`` uses for an SQLAlchemy model class.

    Note also that it is required to set ``unknown`` to ``'preserve'`` in the
    ``__colanderalchemy_config__`` dictionary.

    Example usage

    .. code-block:: python

        from c2cgeoform import models
        from c2cgeoform.ext import deform_ext

        class Photo(models.FileData, Base):
            __tablename__ = 'photo'
            __colanderalchemy_config__ = {
                'title': _('Photo'),
                'unknown': 'preserve',
                'widget': deform_ext.FileUploadWidget(file_upload_temp_store)
            }
            permission_id = Column(Integer, ForeignKey('excavations.id'))

    **Attributes/Arguments**

    id_field (default to "id")
        Name of field to pass to get_url.

    get_url (optional)
        A callback function `function(request, id) -> string` which returns
        the URL to get the file. Example usage:

        .. code-block:: python

            'widget': deform_ext.FileUploadWidget(
                _file_upload_temp_store,
                id_field="id",
                get_url=lambda request, id: request.route_url('file', id=id)
            )
    """

    id_field = "id"
    request: pyramid.request.Request | None = None
    tmpstore: FileUploadTempStore | None = None

    def __init__(
        self,
        get_url: Callable[[pyramid.request.Request, JSON], str] | None = None,
        **kw: Any,
    ) -> None:
        DeformFileUploadWidget.__init__(self, None, **kw)
        self.get_url = get_url

    def populate(self, session: sqlalchemy.orm.Session, request: pyramid.request.Request) -> None:
        del session  # unused
        self.request = request
        self.tmpstore = FileUploadTempStore(request.session)

    def serialize(self, field: deform.field.Field, cstruct: JSONDict, **kw: Any) -> str:
        if cstruct in (colander.null, None):
            cstruct = {}
        kw["url"] = None
        if "uid" not in cstruct and self.id_field in cstruct:
            cstruct["uid"] = cstruct[self.id_field]
            if cstruct[self.id_field] != colander.null and self.get_url:
                kw["url"] = self.get_url(self.request, cstruct[self.id_field])
        if cstruct.get("filename") == colander.null:
            cstruct["filename"] = ""
        return DeformFileUploadWidget.serialize(self, field, cstruct, **kw)  # type: ignore[no-any-return]

    def deserialize(self, field: deform.field.Field, pstruct: str) -> Any:
        value = DeformFileUploadWidget.deserialize(self, field, pstruct)
        if value != colander.null and "fp" in value:
            value["data"] = value.pop("fp")
        return value


class RelationSelectMapWidget(Widget):  # type: ignore[misc]
    """
    Relation select map Deform widget.

    A Deform widget to select an item on a map. From the idea, this widget
    is similar to the ``RelationSelectWidget``, but instead of a select-box
    a map is shown.

    Example usage

    .. code-block:: python

        bus_stop = Column(Integer, ForeignKey('bus_stops.id'), info={
            'colanderalchemy': {
                'title': 'Bus stop',
                'widget': deform_ext.RelationSelectMapWidget(
                    label_field='name', url='/bus_stops'
                )
            }})

    The user is responsible for providing a web-service under the given URL,
    which returns a list of features as GeoJSON. The features must contain the
    two properties specified with `id_field` and `label_field`. The geometries
    are expected to use the CRS `EPSG:4326`.

    To customize the map, the template file `map_select.pt` has to be
    overwritten.

    **Attributes/Arguments**

    url (required)
        The URL to the web-service which returns the GeoJSON features or a
        callback function `function(request: pyramid.request.Request) -> string` which returns the URL
        to the web-service. Example usage:

        .. code-block:: python

            'widget': deform_ext.RelationSelectMapWidget(
                label_field='name',
                url=lambda request: request.route_url('bus_stops')
            )

    label_field
        The property of the GeoJSON features that is used as label.
        Default: ``label``.

    """

    requirements = (
        ("openlayers", "3.0.0"),
        ("c2cgeoform.deform_map", None),
    )

    def __init__(
        self,
        url: str | Callable[[pyramid.request.Request], str],
        label_field: str = "label",
        **kw: Any,
    ) -> None:
        Widget.__init__(self, **kw)
        self.label_field = label_field
        self.get_url: Callable[[pyramid.request.Request], str] = url if callable(url) else lambda request: url
        self.url = None

    def populate(self, session: sqlalchemy.orm.Session, request: pyramid.request.Request) -> None:
        del session  # unused

        if self.url is None:
            self.url = self.get_url(request)  # type: ignore[assignment]

    def serialize(self, field: deform.field.Field, cstruct: str, readonly: bool = False, **kw: Any) -> str:
        if cstruct is colander.null:
            cstruct = ""
        values = self.get_template_values(field, cstruct, kw)
        # make `_` available in template for i18n messages
        values["_"] = TranslationStringFactory("c2cgeoform")
        values["widget_config"] = json.dumps(
            {"labelField": self.label_field, "url": self.url, "readonly": readonly},
        )
        return field.renderer("map_select", **values)  # type: ignore[no-any-return]

    def deserialize(self, field: deform.field.Field, pstruct: str) -> str:
        del field  # Unused
        return pstruct


class RelationSearchWidget(Widget):  # type: ignore[misc]
    """
    Relation search Deform widget.

    A Deform widget to select an item via a search field. This widget is
    similar to the ``RelationSelectWidget``, but instead of a select-box
    a Twitter Typeahead search field is shown.

    Example usage:

    .. code-block:: python

        address_id = Column(Integer, ForeignKey('address.id'), info={
            'colanderalchemy': {
                'title': _('Address'),
                'widget': deform_ext.RelationSearchWidget(
                    url=lambda request: request.route_url('addresses'),
                    model=Address,
                    min_length=1,
                    id_field='id',
                    label_field='label'
                )
            }})

    The user is responsible for providing a web-service at the given URL.  The
    web service should expect requests of the form ``?term=<search_terms>``.
    And it should return responses of this form:

    .. code-block:: json

        [{"id": 0, "label": "foo"}, {"id": 1, "label": "bar"}]

    The name of the id and label keys are configurable. See below.

    **Attributes/arguments**

    url (required)
        The search service URL, or a function that takes a request a return the
        search service URL.

    model (required)
        The SQLAlchemy model class associated to the linked table.

    min_length
        The minimum character length needed before suggestions start getting
        rendered. Default: ``1``.

    id_field
        The name of the "id" property in JSON responses. Default: ``"id"``.

    label_field
        The name of the "label" property in JSON responses.
        Default: ``"label"``.

    limit
        The maximum number of suggestions. Default: 8.

    """

    id_field = "id"
    label_field = "label"
    limit = 8
    min_length = 1
    readonly_template = "readonly/textinput"
    strip = True
    template = "search"
    requirements = (("typeahead", "0.10.5"),)

    def __init__(self, url: str | Callable[[pyramid.request.Request], str], **kw: Any) -> None:
        Widget.__init__(self, **kw)
        self.get_url: Callable[[pyramid.request.Request], str] = url if callable(url) else lambda request: url
        self.url = None
        self.session: sqlalchemy.orm.Session | None = None

    def populate(self, session: sqlalchemy.orm.Session, request: pyramid.request.Request) -> None:
        if self.url is None:
            self.url = self.get_url(request)  # type: ignore[assignment]
        self.session = session

    def serialize(self, field: deform.field.Field, cstruct: str, **kw: Any) -> str:
        if cstruct in (colander.null, None):
            cstruct = ""
            label = ""
        else:
            assert self.session is not None
            obj = self.session.query(self.model).get(cstruct)  # pylint: disable=no-member
            label = getattr(obj, kw.get("label_field", self.label_field))

        kw["label"] = label

        options = {
            "idField": kw.pop("id_field", self.id_field),
            "labelField": kw.pop("label_field", self.label_field),
        }
        kw["options"] = json.dumps(options)

        bloodhound_options = {"limit": kw.pop("limit", self.limit), "remote": f"{self.url}?term=%QUERY"}
        kw["bloodhound_options"] = json.dumps(bloodhound_options)

        typeahead_options = {"minLength": kw.pop("min_length", self.min_length)}
        kw["typeahead_options"] = json.dumps(typeahead_options)

        readonly = kw.get("readonly", self.readonly)

        # If "readonly" is set then deform's readonly "textinput" template will
        # be used. That template will display the value set in "cstruct" so we
        # just set cstruct to label here.
        cstruct = label if readonly else cstruct

        tmpl_values = self.get_template_values(field, cstruct, kw)
        template = self.readonly_template if readonly else self.template

        return field.renderer(template, **tmpl_values)  # type: ignore[no-any-return]

    def deserialize(self, field: deform.field.Field, pstruct: str) -> colander._null | str | None:
        if pstruct is colander.null:
            return colander.null
        if not isinstance(pstruct, string_types):
            raise Invalid(field.schema, "Pstruct is not a string")
        if self.strip:
            pstruct = pstruct.strip()
            if not pstruct:
                return colander.null
            return pstruct
        return None


class RecaptchaWidget(MappingWidget):  # type: ignore[misc]
    """
    A Deform widget for Google reCaptcha.

    In `c2cgeoform` this widget can be used by setting the `show_captcha`
    flag when calling `register_schema()`.

    Example usage:

    .. code-block:: python

        register_schema(
            'comment', model.Comment, show_confirmation=False,
            show_captcha=True,
            recaptcha_public_key=settings.get('recaptcha_public_key'),
            recaptcha_private_key=settings.get('recaptcha_private_key'))

    **Attributes/arguments**

    public_key (required)
        The Google reCaptcha site key.

    private_key (required)
        The Google reCaptcha secret key.

    """

    template = "recaptcha"
    readonly_template = "recaptcha"
    url = "https://www.google.com/recaptcha/api/siteverify"
    request: pyramid.request.Request | None = None

    def populate(self, session: sqlalchemy.orm.Session, request: pyramid.request.Request) -> None:
        del session  # unused
        self.request = request

    def serialize(self, field: deform.field.Field, cstruct: JSONDict, **kw: Any) -> MappingWidget:
        assert self.request is not None
        kw.update(
            {
                "public_key": self.public_key,  # pylint: disable=no-member
                "locale_name": self.request.locale_name,
            },
        )
        return MappingWidget.serialize(self, field, cstruct, **kw)

    def deserialize(self, field: deform.field.Field, pstruct: JSONDict) -> str | colander._null:
        if pstruct is colander.null:
            return colander.null

        # get the verification token that is inserted into a hidden input
        # field created by the reCaptcha script. the value is available in
        # `pstruct` because we are inheriting from `MappingWidget`.
        response = pstruct.get("g-recaptcha-response") or ""
        if not response:
            raise Invalid(field.schema, _("Please verify that you are a human!"), pstruct)
        assert self.request is not None
        remoteip = self.request.remote_addr

        try:
            resp = requests.get(
                self.url,
                params={
                    "secret": self.private_key,  # pylint: disable=no-member
                    "response": response,
                    "remoteip": remoteip,
                },
                timeout=30,
            )
        except requests.exceptions.RequestException as exception:
            log.exception("reCaptcha connection problem")
            raise Invalid(field.schema, _("Connection problem"), pstruct) from exception
        if resp.status_code >= 500:
            log.error("reCaptcha connection problem (%i): %s", resp.status_code, resp.text)
            raise Invalid(field.schema, _("Connection problem"), pstruct)

        error_msg = _("Verification has failed")
        if resp.status_code != 200:
            log.error("reCaptcha validation error: %s", resp.status_code)
            raise Invalid(field.schema, error_msg, pstruct)

        data = resp.json()
        if not data["success"]:
            error_reason = ""
            if "error-codes" in data:
                error_reason = ",".join(data["error-codes"])
            log.error("reCaptcha validation error: %s", error_reason)
            raise Invalid(field.schema, error_msg, pstruct)

        return pstruct



from functools import partial
from itertools import groupby
from unittest import TestCase
from unittest.mock import Mock

import pytest
from bs4 import BeautifulSoup
from pyramid.httpexceptions import HTTPFound, HTTPNotFound

from c2cgeoform.models import DBSession
from c2cgeoform.schema import GeoFormSchemaNode
from c2cgeoform.tests import DatabaseTestCase
from c2cgeoform.tests.models_test import Person, Tag
from c2cgeoform.views.abstract_views import AbstractViews, ListField

_list_field = partial(ListField, Person)


def get_rendered_form(response):
    return response["form"].render(*response["form_render_args"], **response["form_render_kwargs"])


class TestListField(TestCase):
    def test_title_default_to_attr_key(self):
        assert "id" == ListField(Tag, "id").label()


class ConcreteViews(AbstractViews):
    _model = Person
    _id_field = "id"
    _list_fields = [_list_field("name", label="Name"), _list_field("first_name")]
    _base_schema = GeoFormSchemaNode(Person, title="Person")


class TestAbstractViews(DatabaseTestCase):
    def _add_test_persons(self):
        self.person1 = Person(name="Smith", first_name="Peter")
        DBSession.add(self.person1)
        DBSession.add(Person(name="Wayne", first_name="John"))
        DBSession.add(Person(name="Elda", first_name="Hasbrouck"))
        DBSession.add(Person(name="Lashaun", first_name="Brasel"))
        DBSession.add(Person(name="Lashawna", first_name="Ashford"))
        DBSession.add(Person(name="Lesha", first_name="Snellgrove"))
        DBSession.add(Person(name="Sulema", first_name="Page"))
        DBSession.add(Person(name="Derek", first_name="Boroughs"))
        DBSession.add(Person(name="Odis", first_name="Bateman"))
        DBSession.add(Person(name="Venetta", first_name="Briganti"))
        DBSession.add(Person(name="Monte", first_name="Quill"))
        DBSession.add(Person(name="Daniel", first_name="Ruth"))
        DBSession.add(Person(name="Eloise", first_name="Hellickson"))
        DBSession.add(Person(name="Hee", first_name="Deloney"))
        DBSession.add(Person(name="Sharee", first_name="Warf"))
        DBSession.add(Person(name="Delpha", first_name="Philip"))
        DBSession.add(Person(name="Claudio", first_name="Campfield"))
        DBSession.add(Person(name="Janessa", first_name="Beatty"))
        DBSession.add(Person(name="Hollis", first_name="Richmond"))
        DBSession.add(Person(name="Karoline", first_name="Carew"))
        DBSession.add(Person(name="Bess", first_name="Papp"))
        DBSession.add(Person(name="Vada", first_name="Infantino"))
        DBSession.flush()

    def test_index(self):
        self.request.route_url = Mock(return_value="person/new")
        views = ConcreteViews(self.request)
        response = views.index()
        assert "list_fields" in response

    def test_grid(self):
        self.request.route_url = Mock(return_value="person/1")
        self._add_test_persons()

        self.request.params["offset"] = "0"
        self.request.params["limit"] = "5"

        views = ConcreteViews(self.request)
        response = views.grid()

        assert 22 == response["total"]

        rows = response["rows"]
        assert 5 == len(rows)
        assert "_id_" in rows[0]
        assert "Smith" == rows[0]["name"]
        assert "person/1" == rows[0]["actions"]["dblclick"]
        call_list = self.request.route_url.call_args_list
        idx = [call[1]["id"] for call in call_list]
        grouped_by_id = [len(list(cgen)) for dummy, cgen in groupby(idx)]
        grouped_by_count = [len(list(cgen)) for dummy, cgen in groupby(grouped_by_id)]
        assert 1 == len(grouped_by_count)
        assert 5 == grouped_by_count[0]

    def test_grid_without_parameters(self):
        self.request.route_url = Mock(return_value="person/1")
        self._add_test_persons()
        views = ConcreteViews(self.request)
        response = views.grid()
        assert 22 == response["total"]

    def test_grid_with_nan(self):
        self.request.route_url = Mock(return_value="person/1")
        self._add_test_persons()
        self.request.params["offset"] = "NaN"
        self.request.params["limit"] = "NaN"
        views = ConcreteViews(self.request)
        response = views.grid()
        assert 22 == response["total"]

    def test_new_get(self):
        self.request.matched_route = Mock(name="c2cgeoform_item")
        self.request.route_url = Mock(return_value="person/new/edit")

        self.request.matchdict = {"id": "new"}
        views = ConcreteViews(self.request)
        response = views.edit()

        assert "form" in response
        assert "deform_dependencies" in response

    def test_new_post_validation_error(self):
        self.request.matched_route = Mock(name="c2cgeoform_item")
        self.request.route_url = Mock(return_value="person/new/save")

        self.request.matchdict = {"id": "new"}
        self.request.POST["first_name"] = "arnaud"
        self.request.POST["age"] = "37"

        views = ConcreteViews(self.request)
        response = views.save()

        assert "form" in response
        assert "deform_dependencies" in response

    def test_new_post_success(self):
        self.request.matched_route = Mock(name="person_action")
        self.request.route_url = Mock(return_value="person/new/save")
        self.request.dbsession.merge = Mock(
            return_value=Person(name="...", first_name="..."), age="...", id=7869
        )
        self.request.dbsession.flush = Mock()

        self.request.matchdict = {"id": "new"}
        self.request.POST["name"] = "morvan"
        self.request.POST["first_name"] = "arnaud"
        self.request.POST["age"] = "37"

        views = ConcreteViews(self.request)
        response = views.save()

        assert isinstance(response, HTTPFound)

        class Matcher:
            def __eq__(self, other):
                return other.name == "morvan" and other.first_name == "arnaud" and other.age == 37

        self.request.dbsession.merge.assert_called_with(Matcher())
        self.request.dbsession.flush.assert_called_once_with()

    def test_edit_get_not_found(self):
        self.request.matched_route = Mock(name="person_action")
        self.request.matchdict = {"id": 99999}
        self.request.route_url = Mock(return_value="person/99999")

        views = ConcreteViews(self.request)
        with pytest.raises(HTTPNotFound):
            views.edit()

    def test_edit_get_success(self):
        self._add_test_persons()
        self.request.matched_route = Mock(name="person_action")
        self.request.matchdict = {"id": self.person1.id}
        self.request.route_url = Mock(return_value="person/1")

        views = ConcreteViews(self.request)
        response = views.edit()

        assert "form" in response
        assert "deform_dependencies" in response

    def test_edit_post_notfound(self):
        self.request.matched_route = Mock(name="person_action")
        self.request.matchdict = {"id": 99999}
        self.request.route_url = Mock(return_value="person/99999")
        self.request.method = "POST"

        views = ConcreteViews(self.request)
        with pytest.raises(HTTPNotFound):
            views.save()

    def test_duplicate_get(self):
        self._add_test_persons()
        self.request.matched_route = Mock(name="c2cgeoform_item_action")
        self.request.matchdict = {"id": self.person1.id}
        self.request.route_url = Mock(return_value="c2cgeoform_item/new")
        self.request.method = "GET"

        views = ConcreteViews(self.request)
        response = views.duplicate()

        assert "form" in response
        assert "deform_dependencies" in response
        form = BeautifulSoup(get_rendered_form(response), "html.parser")
        # self.assertEqual('', form.select_one('form').attrs['action'])
        assert "" == form.select_one("input[name=id]").attrs["value"]
        assert self.person1.name == form.select_one("input[name=name]").attrs["value"]
        assert self.person1.first_name == form.select_one("input[name=first_name]").attrs["value"]
        assert "" == form.select_one("input[name=age]").attrs["value"]

    def test_copy_and_discard_excluded(self):
        self._add_test_persons()
        self.request.matched_route = Mock(name="c2cgeoform_item_action")
        self.request.matchdict = {"id": self.person1.id}
        self.request.route_url = Mock(return_value="c2cgeoform_item/new")
        self.request.method = "GET"
        person2 = self.request.dbsession.query(Person).filter(Person.name == "Wayne").one_or_none()
        views = ConcreteViews(self.request)

        response = views.copy(person2)
        form = BeautifulSoup(get_rendered_form(response), "html.parser")

        assert person2.name == form.select_one("input[name=name]").attrs["value"]
        assert person2.first_name == form.select_one("input[name=first_name]").attrs["value"]
        assert "" == form.select_one("input[name=age]").attrs["value"]

        response = views.copy(person2, ["first_name"])
        form = BeautifulSoup(get_rendered_form(response), "html.parser")

        assert person2.name == form.select_one("input[name=name]").attrs["value"]
        assert "" == form.select_one("input[name=first_name]").attrs["value"]
        assert "" == form.select_one("input[name=age]").attrs["value"]

    def test_delete(self):
        dbsession = self.request.dbsession

        self._add_test_persons()
        self.request.matched_route = Mock(name="person_action")
        self.request.matchdict = {"id": self.person1.id}
        self.request.route_url = Mock(return_value="person")
        dbsession.delete = Mock()
        flush_which_has_to_be_back_for_teardown = dbsession.flush
        try:
            dbsession.flush = Mock()

            views = ConcreteViews(self.request)
            response = views.delete()

            class Matcher:
                def __init__(self, person) -> None:
                    self.id = person.id

                def __eq__(self, other):
                    return other.id == self.id

            assert True is response["success"]
            assert "person" == response["redirect"]
            dbsession.delete.assert_called_once_with(Matcher(self.person1))
            dbsession.flush.assert_called_once_with()
        finally:
            dbsession.flush = flush_which_has_to_be_back_for_teardown
